#include "partition.h"
#include "mesh.h"
extern Mesh init_mesh;
extern Mesh contracted_mesh;
#include "cluster.h"
extern PCluster* pcluster;
#include <vector>
extern int valid_clusters;
extern int * valid_mapping;
extern string in_file;
extern string out_file;
#include <fstream>

Partition::Partition():
	__cov(init_mesh.mesh_face.size(), CovDet()),
	initial_cov(init_mesh.mesh_face.size(), CovDet())
{
	D = 3;
	valid_clusters = init_mesh.mesh_face.size();
	cluster_energy.clear();
	swap_started = false;
	info_set.clear();

	vector<face_pair> s;
	initface_neighbor_links = vector<vector<face_pair>>(init_mesh.mesh_face.size(), s);
	cluster_neighbor_links = vector<vector<domain_pair>>(init_mesh.mesh_face.size(), s);
	set<FaceID> s2;
	cluster_boundary_faces = vector<set<FaceID>>(init_mesh.mesh_face.size(), s2);
	vector<face_pair*> s3;
	face_neighbor_links= vector<vector<face_pair*>>(init_mesh.mesh_face.size(), s3);
}

Partition::~Partition()
{

}

void Partition::init_cov()
{
	pcluster = new PCluster(valid_clusters);
	collect_triangles_covariance();

	collect_face_pair();
	//make a copy of the covariance and face neighbor information
	for(uint j=0; j<cov_count(); j++)
	{
		init_covariance(j) = covariance(j);
		for (int k=0; k<face_neighbor_links[j].size(); k++)
		{			
			face_pair *info = face_neighbor_links[j][k];
			if (info->f1 < info->f2)
			{
				initface_neighbor_links[info->f1].push_back(*info);
				initface_neighbor_links[info->f2].push_back(*info);
			}			
		}
	}
}

void Partition::collect_triangles_covariance()
{
	for (FaceID i=0; i<cov_count(); i++)
	{
		int v1 = init_mesh.mesh_face[i].x;
		int v2 = init_mesh.mesh_face[i].y;
		int v3 = init_mesh.mesh_face[i].z;
		Vector3d coor1 = Vector3d(init_mesh.mesh_node[v1].x, init_mesh.mesh_node[v1].y, init_mesh.mesh_node[v1].z);
		Vector3d coor2 = Vector3d(init_mesh.mesh_node[v2].x, init_mesh.mesh_node[v2].y, init_mesh.mesh_node[v2].z);
		Vector3d coor3 = Vector3d(init_mesh.mesh_node[v3].x, init_mesh.mesh_node[v3].y, init_mesh.mesh_node[v3].z);
		CovDet cov(coor1, coor2, coor3);
		covariance(i) = cov;
	}
}

void Partition::collect_face_pair()
{
	init_mesh.initMeshInfo();
	for (FaceID i=0; i<init_mesh.mesh_face.size(); i++)
	{
		vector<int> face_neighbors = init_mesh.getFaceNeighbors(i);
		for (vector<int>::iterator it = face_neighbors.begin(); it != face_neighbors.end(); it++)
		{
			if (i < *it)
			{
				create_pair(i, *it);
			}			
		}
	}
}

void Partition::create_pair(FaceID j, FaceID k)
{
	face_pair *info = new face_pair();
	info->f1 = j;
	info->f2 = k;
	compute_pair_info(info);

	face_neighbor_links[j].push_back(info);
	face_neighbor_links[k].push_back(info);
}

void Partition::compute_pair_info(face_pair* info)
{
	compute_energy(info);

	finalize_pair_update(info);
}

void Partition::compute_energy(face_pair *info)
{
	FaceID i=info->f1, j=info->f2;

	CovDet &Ci=covariance(i), &Cj=covariance(j);
	CovDet C=Ci;  C+=Cj;

	double err = C.energy() - Ci.energy() - Cj.energy();
	info->heap_key(-err);
}

void Partition::finalize_pair_update(face_pair *info)
{
	if( info->is_in_heap() )
		heap.update(info);
	else
		heap.insert(info);
}

bool Partition::decimate(uint target)
{
	while( valid_clusters > target )
	{
		face_pair *info = (face_pair *)heap.extract();
		if( !info ) 
		{
			return false;
		}
			
		FaceID v1=info->f1, v2=info->f2;

		if( contracted_mesh.face_is_valid(v1) && contracted_mesh.face_is_valid(v2) )
		{
			apply_contraction(info);
			valid_clusters--;
			delete info;
		}			
	}

	valid_mapping = new int[valid_clusters];
	int n_domain = 0;
	for(FaceID j=0; j<pcluster->size; j++)
	{
		if (pcluster->face_clusters[j].size() != 0)
		{
			valid_mapping[n_domain] = j;
			n_domain ++;
		} 
	}
	assert(n_domain == valid_clusters);
	return true;
}

void Partition::apply_contraction(face_pair *info)
{
	int f1 = info->f1;
	int f2 = info->f2;
	pcluster->mergeDomain(f1, f2);
	covariance(f1) += covariance(f2);
	// Move v1 to new position
	contracted_mesh.set_position(f1, covariance(f1).centroid(0), covariance(f1).centroid(1), covariance(f1).centroid(2));
	//invalidate v2
	contracted_mesh.mark_invalid(f2);

	update_valid_pairs(info);
	for(uint i=0; i<face_neighbor_links[f1].size(); i++)
		compute_pair_info(face_neighbor_links[f1][i]);
}

void Partition::update_valid_pairs(face_pair *info)
{
	//find all the possible neighbors
	FaceID f1 = info->f1, f2 = info->f2;
	set<FaceID> neighbors;
	int n_neighbors = find_contracted_neighbors(f1, neighbors);
	uint j;

	//update the heap
	for(int i=0; i<face_neighbor_links[f1].size(); i++)
	{
		face_pair *e = face_neighbor_links[f1][i];
		FaceID u = (e->f1==f1)?e->f2:e->f1;
		heap.remove(e);
		erase_pair_counterpart(u,e);
		if(u != f2)
			delete e;
	}
	for(int i=0; i<face_neighbor_links[f2].size(); i++)
	{
		face_pair *e = face_neighbor_links[f2][i];
		FaceID u = (e->f1==f2)?e->f2:e->f1;
		heap.remove(e);
		erase_pair_counterpart(u,e);
		if(u != f1)
			delete e;
	}

	face_neighbor_links[f1].clear();
	face_neighbor_links[f2].clear();
	for (set<FaceID>::iterator it = neighbors.begin(); it != neighbors.end(); it++)
	{
		face_pair * info_to_add =  new face_pair();
		info_to_add->f1 = f1;
		info_to_add->f2 = *it;
		face_neighbor_links[f1].push_back(info_to_add);
		face_neighbor_links[*it].push_back(info_to_add);	
	}
}

int Partition::find_contracted_neighbors(FaceID v1, set<FaceID> & neighbors)
{
	neighbors.clear();

	//go down to all the point of cluster v1, find all the neighbors of all the points, if the neighbor is not in v1, the belonging domain is a potential neighbor
	for (set<FaceID>::iterator it = pcluster->face_clusters[v1].begin(); it != pcluster->face_clusters[v1].end(); it++)
	{
		for (int neighbor_count = 0; neighbor_count< initface_neighbor_links[*it].size(); neighbor_count ++)
		{
			face_pair e = initface_neighbor_links[*it][neighbor_count];
			FaceID neighbor_point = (e.f1==*it)?e.f2:e.f1;
			if (pcluster->cluster_belong[neighbor_point] != v1)
				neighbors.insert(pcluster->cluster_belong[neighbor_point]);
		}
	}
	
	return neighbors.size();
}

void Partition::erase_pair_counterpart(ClusterID u, face_pair *info)
{
	for (int i =0; i<face_neighbor_links[u].size(); i++)
	{
		if(face_neighbor_links[u][i] == info)
			face_neighbor_links[u].erase(face_neighbor_links[u].begin()+i);
	}
}

void Partition::init_face_swap()
{
	prepare_space();
	collect_cluster_neighbor();

	for (int i=0; i< valid_clusters; i++)
		update_cluster_information(i);
	for (int i=0; i< valid_clusters; i++)
		collect_cluster_swaps(i);

	//print_total_energy();
}

void Partition::prepare_space()
{
	cluster_energy = vector<double>(valid_clusters, 0);
}

void Partition::collect_cluster_neighbor()
{
	//delete all the neighboring information
	for (int i=0; i< valid_clusters; i++)
	{
		cluster_neighbor_links[valid_mapping[i]].clear();
		cluster_boundary_faces[valid_mapping[i]].clear();
	}
	//go down to all the point of cluster i, find all the neighbors of all the faces, if the neighbor is not in cluster i, the belonging cluster is the neighbor cluster
	for (int i=0; i< valid_clusters; i++)
	{
		set<int> neighbors;
		for (set<FaceID>::iterator it = pcluster->face_clusters[valid_mapping[i]].begin(); it != pcluster->face_clusters[valid_mapping[i]].end(); it++)
		{
			for (int neighbor_count = 0; neighbor_count< initface_neighbor_links[*it].size(); neighbor_count ++)
			{
				face_pair e = initface_neighbor_links[*it][neighbor_count];
				FaceID neighbor_face = (e.f1==*it)?e.f2:e.f1;

				//if it has a neighbor that not belong to v1, then it's at the boundary
				if (valid_mapping[i] != pcluster->cluster_belong[neighbor_face])
					cluster_boundary_faces[valid_mapping[i]].insert(*it);
				//if the neighbor is not in cluster i, the belonging cluster is the neighbor cluster
				if (valid_mapping[i] < pcluster->cluster_belong[neighbor_face])
					neighbors.insert(pcluster->cluster_belong[neighbor_face]);
			}
		}

		for (set<int>::iterator it = neighbors.begin(); it != neighbors.end(); it++)
		{
			domain_pair info;
			info.f1 = valid_mapping[i];
			info.f2 = *it;
			cluster_neighbor_links[valid_mapping[i]].push_back(info);
			cluster_neighbor_links[*it].push_back(info);
		}
	}
}

void Partition::update_cluster_information(ClusterID i)
{
	CovDet &Cj=covariance(valid_mapping[i]);
	contracted_mesh.cluster_pos[valid_mapping[i]].x = Cj.centroid(0);
	contracted_mesh.cluster_pos[valid_mapping[i]].y = Cj.centroid(1);
	contracted_mesh.cluster_pos[valid_mapping[i]].z = Cj.centroid(2);
	cluster_energy[i] = Cj.energy();
}

void Partition::collect_cluster_swaps(ClusterID i)
{
	for (set<FaceID>::iterator face_it = cluster_boundary_faces[valid_mapping[i]].begin(); face_it != cluster_boundary_faces[valid_mapping[i]].end(); face_it++)
	{
		ClusterID cluster_closest = valid_mapping[i];
		double energy_saved_closest = 0;
		face_swap *info = new face_swap();

		//compute which domain it should belong to
		for (int neighbor_count = 0; neighbor_count< initface_neighbor_links[*face_it].size(); neighbor_count ++)
		{
			face_pair e = initface_neighbor_links[*face_it][neighbor_count];
			FaceID neighbor_face = (e.f1 ==*face_it ? e.f2 : e.f1);
			ClusterID cluster_neighbor = pcluster->cluster_belong[neighbor_face];
			if (valid_mapping[i] != cluster_neighbor)
			{
				double energy_saved = calc_energy_delta(valid_mapping[i], cluster_neighbor, *face_it, info);
				if (energy_saved > energy_saved_closest)
				{
					cluster_closest = cluster_neighbor;
					energy_saved_closest = energy_saved;
				}
			}		
		}

		// insert
		if (energy_saved_closest > 0)
			store_swap(energy_saved_closest, valid_mapping[i], cluster_closest, *face_it, info);
		else
			delete info;
	}		
}

double Partition::calc_energy_delta(ClusterID from, ClusterID to, FaceID face_id, face_swap *info)
{
	const CovDet &Ci=covariance(from), &Cj=covariance(to), &C_init = init_covariance(face_id);
	CovDet C_from = Ci;
	CovDet C_to = Cj;
	double energy_before = cluster_energy[find_cluster_id(from)] + cluster_energy[find_cluster_id(to)];
	C_from -= C_init;
	C_to += C_init;

	double energy_from = C_from.energy();
	double energy_to = C_to.energy();
	double energy_after = energy_from + energy_to;
	return energy_before - energy_after;
}

void Partition::store_swap(float energy_saved, ClusterID from, ClusterID to, FaceID face_id, face_swap *info)
{
	info->from_id = from;
	info->to_id = to;
	info->face_id = face_id;
	info->heap_key(energy_saved);

	info_set.push_back(info);
}

bool Partition::swap_face_once()
{
	set<ClusterID> recollect_domain;//the domains need update	
	apply_swap_set(info_set, recollect_domain);
	collect_cluster_neighbor();
	for (vector<face_swap *>::iterator it = info_set.begin(); it != info_set.end(); it++)
		delete *it;
	info_set.clear();
	for (set<ClusterID>::iterator seed_it = recollect_domain.begin(); seed_it != recollect_domain.end(); seed_it++)
	{
		collect_cluster_swaps(find_cluster_id(*seed_it));
	}

	double domain_energy = print_total_energy();
	if (!swap_started)
	{
		swap_started = true;
		energy_previous_iter = domain_energy;
		return true;
	}
	else
	{
		if ((energy_previous_iter - domain_energy)/energy_previous_iter > energy_convergence)
		{
			energy_previous_iter = domain_energy;
			return true;
		}
		else
			return false;
	}
}

void Partition::apply_swap_set(vector<face_swap *> info_set, set<ClusterID>& recollect_domain)
{
	set<ClusterID> update_domain;// the domain sets that need to update information
	update_domain.clear();
	for (vector<face_swap *>::iterator info_it = info_set.begin(); info_it != info_set.end(); info_it++)
	{
		apply_swap(*info_it, update_domain);
	}
	//update information
	for(set<ClusterID>::iterator seed_it = update_domain.begin(); seed_it != update_domain.end(); seed_it++)
		update_cluster_information(find_cluster_id(*seed_it));

	//domain's neighbors also need get updated
	recollect_domain.clear();
	for(set<ClusterID>::iterator seed_it = update_domain.begin(); seed_it != update_domain.end(); seed_it++)
	{
		for(uint j=0; j<cluster_neighbor_links[*seed_it].size(); j++)
		{
			recollect_domain.insert(cluster_neighbor_links[*seed_it][j].f1);
			recollect_domain.insert(cluster_neighbor_links[*seed_it][j].f2);
		}
	}
}

void Partition::apply_swap(face_swap* info, set<ClusterID>& update_domain)
{
	pcluster->face_clusters[info->from_id].erase(info->face_id);
	pcluster->face_clusters[info->to_id].insert(info->face_id);
	pcluster->cluster_belong[info->face_id]=info->to_id;

	//update covariance of v1 and v2 
	covariance(info->from_id) -= init_covariance(info->face_id);
	covariance(info->to_id) += init_covariance(info->face_id);

	//collect the domains that need update
	update_domain.insert(info->from_id);
	update_domain.insert(info->to_id);
}

void Partition::output_partition_ply()
{
	string save_file = out_file.substr(0, out_file.find_last_of(".")) + "_partition.ply";
	cout << "output partition result to " << save_file.c_str() << endl;
	ofstream of(save_file.c_str());
	of << "ply" << endl;
	of << "format ascii 1.0" << endl;
	of << "element vertex " << init_mesh.mesh_node.size() << endl;
	of << "property float x" << endl;
	of << "property float y" << endl;
	of << "property float z" << endl;
	of << "element face " << init_mesh.mesh_face.size() << endl;
	of << "property list uchar int vertex_indices" << endl;
	of << "property uchar red" << endl;
	of << "property uchar green" << endl;
	of << "property uchar blue" << endl;
	of << "property uchar alpha" << endl;
	of << "end_header" << endl;

	for (int i=0; i< init_mesh.mesh_node.size(); i++)
	{
		of <<  init_mesh.mesh_node[i].x << '\t' <<  init_mesh.mesh_node[i].y << '\t' <<  init_mesh.mesh_node[i].z << endl;
	}
	for (int i=0; i<init_mesh.mesh_face.size(); i++)
	{
		of << 3 << '\t' << init_mesh.mesh_face[i].x << '\t' <<  init_mesh.mesh_face[i].y << '\t' <<  init_mesh.mesh_face[i].z << '\t';
		of << (int)(pcluster->red[pcluster->cluster_belong[i]]*255) << '\t' << (int)(pcluster->green[pcluster->cluster_belong[i]]*255) << '\t' << (int)(pcluster->blue[pcluster->cluster_belong[i]]*255)
			<< '\t' << 255 << endl;
	}
	of.close();
	cout << "export partition complete!" << endl;
}

int Partition::find_cluster_id(int after_mapping)
{
	int lower_id = 0;
	int higher_id= valid_clusters-1;

	while (higher_id >= lower_id)
	{
		int imid = (higher_id + lower_id)/2;

		// determine which subarray to search
		if (valid_mapping[imid] < after_mapping)
			lower_id = imid + 1;
		else if (valid_mapping[imid] > after_mapping)
			higher_id = imid - 1;
		else
			return imid;
	}
	return -1;
}

float Partition::print_total_energy()
{
	float total_energy = 0;
	for (int i=0; i< valid_clusters; i++)
		total_energy += cluster_energy[i];
	cout << "..";

	return total_energy;
}

vector<int> Partition::find_neighbor_clusters(int i)
{
	vector<int> neighbors;
	for (int j=0; j<cluster_neighbor_links[valid_mapping[i]].size(); j++)
	{
		face_pair e = cluster_neighbor_links[valid_mapping[i]][j];
		ClusterID u = (e.f1==valid_mapping[i])?e.f2:e.f1;
		int neighbor_domain_id = find_cluster_id(u);
		neighbors.push_back(neighbor_domain_id);
	}
	return neighbors;
}