#include "remesh.h"
#include "partition.h"
#include "cluster.h"
#include <fstream>

extern Mesh init_mesh;
extern Mesh contracted_mesh;
extern Partition * mesh_partition;
extern PCluster* pcluster;
extern int valid_clusters;
extern int * valid_mapping;
extern string in_file;
extern string out_file;

Remesh::Remesh()
{
	anchor_nodes.clear();
	nodes.clear();
	nodes_n.clear();
	triangle_faces.clear();
	prox.clear();
	poly_num = 0;
	dijkstra = NULL;

	poly_mesh_ready = false;
}

void Remesh::cleanClusters()
{
	vector<bool> face_visited = vector<bool>(init_mesh.mesh_face.size(), false);

	for (int i=0; i<valid_clusters; i++)
	{
		int cluster_id = valid_mapping[i];

		vector<FaceID> flooding_faces;
		vector<set<FaceID>> sub_clusters;
		vector<set<int>> sub_cluster_neighbors;
		set<FaceID> remain_components = pcluster->face_clusters[cluster_id];
		int ncluster = 0;
		set<FaceID> new_cluster;
		set<int> new_cluster_neighbors;
		//separate each connected components as a sub-cluster
		while (!remain_components.empty())
		{
			ncluster++;
			flooding_faces.clear();
			FaceID starting_face = *(remain_components.begin());
			flooding_faces.push_back(starting_face);		
			remain_components.erase(starting_face);
			face_visited[starting_face] = true;

			new_cluster.clear();
			new_cluster_neighbors.clear();
			new_cluster.insert(starting_face);

			//find the new cluster by flooding
			while(!flooding_faces.empty())
			{
				vector<int>::iterator it = flooding_faces.begin();
				int p_id = *it;
				flooding_faces.erase(it);
				vector<int> neighbor_faces = init_mesh.getFaceNeighbors(p_id);

				for (vector<int>::iterator f_it = neighbor_faces.begin(); f_it != neighbor_faces.end(); f_it++)
				{
					FaceID neighbor_face = *f_it;
					//the neighbors in the set should be marked as connected
					set<FaceID>::iterator cluster_neighbor_it = remain_components.find(neighbor_face);
					if ( cluster_neighbor_it != remain_components.end())
					{
						if ( !face_visited[*cluster_neighbor_it] )
						{
							flooding_faces.push_back(*cluster_neighbor_it);
							face_visited[*cluster_neighbor_it] = true;
							new_cluster.insert(*cluster_neighbor_it);
							remain_components.erase(cluster_neighbor_it);
						}			
					}
					new_cluster_neighbors.insert(pcluster->cluster_belong[neighbor_face]);
				}
			}

			sub_clusters.push_back(new_cluster);
			new_cluster_neighbors.erase(cluster_id);
			sub_cluster_neighbors.push_back(new_cluster_neighbors);
		}

		int largest_component_id = 0;
		int largest_size = sub_clusters[0].size();
		for (int j=0; j< ncluster; j++)
		{
			if (sub_clusters[j].size() > largest_size)
			{
				largest_component_id = j;
				largest_size = sub_clusters[j].size();
			}
		}
		//hold the largest connected component as a cluster
		//update the cluster, covariance matrix, set the centroid as the node pos
		if (sub_cluster_neighbors[largest_component_id].size() >= 3) 
		{
			pcluster->face_clusters[cluster_id] = sub_clusters[largest_component_id];
			CovDet cov = CovDet();
			for(set<FaceID>::iterator f_it = sub_clusters[largest_component_id].begin(); f_it != sub_clusters[largest_component_id].end(); f_it++)
			{
				cov += mesh_partition->init_covariance(*f_it);
				face_visited[*f_it] = false;
			}
			mesh_partition->covariance(cluster_id) = cov;
			contracted_mesh.set_position(cluster_id, cov.centroid(0), cov.centroid(1), cov.centroid(2));
		}
		else
		{
			largest_component_id = -1;
			pcluster->face_clusters[cluster_id].clear();
			mesh_partition->covariance(cluster_id) = CovDet();
		}

		//merge other connected components
		for (int component_id =0; component_id < ncluster; component_id ++)
		{
			if (component_id != largest_component_id)
			{
				CovDet cov_i = CovDet();
				for(set<FaceID>::iterator f_it = sub_clusters[component_id].begin(); f_it != sub_clusters[component_id].end(); f_it++)
				{
					cov_i += mesh_partition->init_covariance(*f_it);			
				}

				int cluster_closest = *(sub_cluster_neighbors[component_id].begin());
				double energy_increased_smallest = 1e20;
				//merge with its neighbors
				CovDet cov_j;
				CovDet C;
				for (set<int>::iterator cluster_it = sub_cluster_neighbors[component_id].begin(); cluster_it != sub_cluster_neighbors[component_id].end(); cluster_it++)
				{
					cov_j=mesh_partition->covariance(*cluster_it);
					C=cov_i;  C+=cov_j;

					double energy_increased = C.energy() - cov_i.energy() - cov_j.energy();
					//assert(energy_increased >= 0);
					if (energy_increased < energy_increased_smallest)
					{
						cluster_closest = *cluster_it;
						energy_increased_smallest = energy_increased;
					}
				}

				//update the cluster, covariance matrix, set the centroid as the node pos
				for(set<FaceID>::iterator f_it = sub_clusters[component_id].begin(); f_it != sub_clusters[component_id].end(); f_it++)
				{
					pcluster->face_clusters[cluster_closest].insert(*f_it);
					pcluster->cluster_belong[*f_it] = cluster_closest;
					face_visited[*f_it] = false;
				}
				mesh_partition->covariance(cluster_closest) += cov_i;
				contracted_mesh.set_position(cluster_closest, mesh_partition->covariance(cluster_closest).centroid(0), mesh_partition->covariance(cluster_closest).centroid(1), mesh_partition->covariance(cluster_closest).centroid(2));
			}
		}
	}

	delete [] valid_mapping;
	valid_mapping = new int[valid_clusters];
	valid_clusters = 0;
	int n_domain = 0;
	for(FaceID j=0; j<pcluster->size; j++)
	{
		if (pcluster->face_clusters[j].size() != 0)
		{
			valid_mapping[valid_clusters] = j;
			valid_clusters ++;
		} 
	}
}

void Remesh::construct_poly_mesh()
{
	find_anchor_vertices();
	project_anchor_vertices();
	init_nodes();
	edges_extraction();
	loops_extraction();
	orient_loops();
	output_poly_mesh();
	poly_mesh_ready = true;
}

void Remesh::find_anchor_vertices()
{
	set<int> neighbor_cluster_set;
	int cluster_count;
	set<Anchor_node, AnchorCompare>::iterator a_it;

	//the clusters that more than 3 clusters meet
	set<set<int>> multi_clusters;
	multi_clusters.clear();

	for (int i=0; i<init_mesh.mesh_node.size(); i++)
	{
			cluster_count = getMeetingClusters(i,neighbor_cluster_set);
			if (cluster_count >=3 || (init_mesh.mesh_node[i].boundary_node && cluster_count >= 2))
			{
				Anchor_node anchor = Anchor_node(neighbor_cluster_set, i);
				if(init_mesh.mesh_node[i].boundary_node)
					anchor.boundary = true;
				a_it = anchor_nodes.find(anchor);
				if (a_it == anchor_nodes.end())
				{
					anchor_nodes.insert(anchor);
				}
				else
				{
					set<int> pid = a_it->init_nodes_candidates;
					pid.insert(i);
					
					anchor = Anchor_node(neighbor_cluster_set, pid);	
					if (a_it->boundary || init_mesh.mesh_node[i].boundary_node)
						anchor.boundary = true;
					anchor_nodes.erase(a_it);
					anchor_nodes.insert(anchor);
				}

				if (cluster_count > 3)
				{
					multi_clusters.insert(neighbor_cluster_set);
				}
			}
	}
	//remove redundant meeting set
	for (set<set<int>>::iterator it = multi_clusters.begin(); it != multi_clusters.end(); it++)
	{
		int s[20];
		int t[20];
		int i=0;
		for (set<int>::iterator s_it = it->begin(); s_it != it->end(); s_it++)
		{
			s[i] = *s_it;
			i++;
		}
		mergeKSubset(anchor_nodes, s, it->size(), 3, t, 0, 0);
	}
}

int Remesh::getMeetingClusters(int nodeid, set<int> & clusterset)
{
	clusterset.clear();
	vector<int> one_ring_faces = init_mesh.node_neighbors[nodeid];
	for (vector<int>::iterator it = one_ring_faces.begin(); it != one_ring_faces.end(); it++)
	{
		clusterset.insert(pcluster->cluster_belong[*it]);
	}
	return clusterset.size();
}

//Backtracking, just show off my programing skills, this function is made to be hard to understand on purpose
void Remesh::mergeKSubset(set<Anchor_node, AnchorCompare>& anchor,int s[],int p,int k, int t[],int q,int r)
{
	if(q==k)
	{
		//merge the candidate of the subset to set
		set<int> set_cluster, sub_cluster;
		for (int i=0; i< p; i++)
		{
			set_cluster.insert(s[i]);
		}
		for (int i=0; i< q; i++)
		{
			sub_cluster.insert(t[i]);
		}
		set<int> sub_pid;
		set<Anchor_node, AnchorCompare>::iterator a_it = anchor.find(Anchor_node(sub_cluster, sub_pid));
		if (a_it != anchor.end())
		{
			sub_pid = a_it->init_nodes_candidates;
			a_it = anchor.find(Anchor_node(set_cluster, sub_pid));
			set<int> original_pid = a_it->init_nodes_candidates;
			anchor.erase(Anchor_node(set_cluster, original_pid));
			for (set<int>::iterator it = sub_pid.begin(); it != sub_pid.end(); it++)
			{
				original_pid.insert(*it);
			}
			anchor.insert(Anchor_node(set_cluster, original_pid));
			anchor.erase(Anchor_node(sub_cluster, sub_pid));
		}
	}
	else
	{
		for(int i=r;i<p;i++)
		{
			t[q]=s[i];
			mergeKSubset(anchor,s,p,k,t,q+1,i+1);
		}
	}
}

void Remesh::project_anchor_vertices()
{
	set<Anchor_node, AnchorCompare> anchor_avg_projection;
	anchor_avg_projection.clear();

	Vector4d plane_equ[20];
	vec3 point_project[20];
	vec3 point_to_center[20];
	for (set<Anchor_node, AnchorCompare>::iterator a_it = anchor_nodes.begin(); a_it != anchor_nodes.end(); a_it++)
	{
		//get the planes
		int ncluster = a_it->meeting_clusters.size();
		set<int> meeting_clusters = a_it->meeting_clusters;
		int m = 0;
		for (set<int>::iterator it = meeting_clusters.begin(); it != meeting_clusters.end(); it++)
		{
			vec3 cluster_n = mesh_partition->covariance(*it).getPlaneNormal();
			vec3 cluster_p = vec3(mesh_partition->covariance(*it).centroid(0), mesh_partition->covariance(*it).centroid(1), mesh_partition->covariance(*it).centroid(2));
			plane_equ[m] = Vector4d(cluster_n[0],cluster_n[1],cluster_n[2],-cluster_n.dot(cluster_p));
			m++;
		}

		//find the closest point
		Vector4d closest_p;
		float closest_dis = 1e20;
		for(set<int>::iterator p_it = a_it->init_nodes_candidates.begin(); p_it != a_it->init_nodes_candidates.end(); p_it++)
		{
			Vector4d point_p = Vector4d(init_mesh.mesh_node[*p_it].x, init_mesh.mesh_node[*p_it].y, init_mesh.mesh_node[*p_it].z, 1);
			float point_dis = 0;
			for (int i=0; i< ncluster; i++)
			{
				float point_plane_dis = point_p.transpose() * plane_equ[i];
				point_dis += abs(point_plane_dis);
			}

			if (point_dis < closest_dis)
			{
				closest_p = point_p;
				closest_dis = point_dis;
				a_it->closest_node = *p_it;
			}
		}

		//project the point on all the planes and average the projections
		m=0;
		for (set<int>::iterator it = meeting_clusters.begin(); it != meeting_clusters.end(); it++)
		{
			vec3 cluster_n = vec3(plane_equ[m](0), plane_equ[m](1), plane_equ[m](2));
			float point_plane_dis = closest_p.transpose() * plane_equ[m];
			point_project[m] = vec3(closest_p(0), closest_p(1), closest_p(2)) - (cluster_n*point_plane_dis);
			m++;
		}

		Anchor_node anchor = *a_it;
		anchor.pos.zero();
		for (int i=0; i< ncluster; i++)
		{
			anchor.pos += point_project[i];
		}
		anchor.pos /= ncluster;
		anchor_avg_projection.insert(anchor);
	}
	anchor_nodes = anchor_avg_projection;
}

void Remesh::init_nodes()
{
	nodes.clear();
	for (set<Anchor_node, AnchorCompare>::iterator it= anchor_nodes.begin(); it != anchor_nodes.end(); it++)
	{
		nodes.push_back(it->pos);
	}

	nodes_n.clear();
	vector<vec3> cluster_n = vector<vec3>(valid_clusters, vec3(0,0,0));
	for (int i=0; i<valid_clusters; i++)
	{
		cluster_n[i] = mesh_partition->covariance(valid_mapping[i]).getPlaneNormal(); //unoriented normal needs to be checked
		vec3 weighted_cluster_normal = vec3(0,0,0);
		for (set<FaceID>::iterator it = pcluster->face_clusters[valid_mapping[i]].begin(); it != pcluster->face_clusters[valid_mapping[i]].end(); it++)
		{
			int v1 = init_mesh.mesh_face[*it].x;
			int v2 = init_mesh.mesh_face[*it].y;
			int v3 = init_mesh.mesh_face[*it].z;

			vec3 p1 = vec3(init_mesh.mesh_node[v1].x, init_mesh.mesh_node[v1].y, init_mesh.mesh_node[v1].z);
			vec3 p2 = vec3(init_mesh.mesh_node[v2].x, init_mesh.mesh_node[v2].y, init_mesh.mesh_node[v2].z);
			vec3 p3 = vec3(init_mesh.mesh_node[v3].x, init_mesh.mesh_node[v3].y, init_mesh.mesh_node[v3].z);
			weighted_cluster_normal += (p2-p1).cross(p3-p2);
		}

		if(cluster_n[i].dot(weighted_cluster_normal) < 0)
			cluster_n[i] = vec3(0,0,0) - cluster_n[i];
	}

	for (set<Anchor_node, AnchorCompare>::iterator it= anchor_nodes.begin(); it != anchor_nodes.end(); it++)
	{
		vec3 mean_cluster_normal = vec3(0,0,0);
		for (set<int>::iterator cluster_it = it->meeting_clusters.begin(); cluster_it != it->meeting_clusters.end(); cluster_it++)
		{
			mean_cluster_normal += cluster_n[mesh_partition->find_cluster_id(*cluster_it)];
		}
		mean_cluster_normal.normalize();
		nodes_n.push_back(mean_cluster_normal);
	}
}

void Remesh::edges_extraction()
{
	prox = vector<Proxy>(valid_clusters, Proxy());
	set<int> belonging_prox;
	int global_id = 0;
	for (set<Anchor_node, AnchorCompare>::iterator it = anchor_nodes.begin(); it != anchor_nodes.end(); it++)
	{	
		for (set<int>::iterator prox_it = it->meeting_clusters.begin(); prox_it != it->meeting_clusters.end(); prox_it++)
		{
			belonging_prox= it->meeting_clusters;
			belonging_prox.erase(*prox_it);
			if(it->boundary)
				belonging_prox.insert(-1);
			Anchor_node anchor(belonging_prox, 0);
			anchor.node_id = global_id;
			anchor.closest_node = it->closest_node;
			int prox_id = mesh_partition->find_cluster_id(*prox_it);
			prox[prox_id].neighboring_cluster_info.push_back(anchor);
		}
		global_id++;
	}

	set<int> nc1, nc2, nc3;
	for (int i=0; i<valid_clusters; i++)
	{
		for (vector<Anchor_node>::iterator node_it1 = prox[i].neighboring_cluster_info.begin(); node_it1 != prox[i].neighboring_cluster_info.end(); node_it1++)
		{
			nc1 = node_it1->meeting_clusters;
			for (vector<Anchor_node>::iterator node_it2 = next(node_it1); node_it2 != prox[i].neighboring_cluster_info.end(); node_it2++)
			{
				nc2 = node_it2->meeting_clusters;
				nc3.clear();
				set_intersection(nc1.begin(), nc1.end(), nc2.begin(), nc2.end(), std::inserter(nc3, nc3.end()));
				if (!nc3.empty())
				{
					prox[i].edges.push_back(Edge(node_it1->node_id, node_it2->node_id));
				}
			}
		}
	}
}

void Remesh::loops_extraction()
{
	for (int i=0; i<valid_clusters; i++)
	{
		if (prox[i].edges.size() >= 3)
		{
			int node_start = prox[i].edges.begin()->v1;
			int node_to_search = prox[i].edges.begin()->v2;
			prox[i].loop.push_back(node_start);
			prox[i].edges.begin()->used_in_loop = true;
			while (node_to_search != node_start)
			{
				for (vector<Edge>::iterator it = prox[i].edges.begin(); it != prox[i].edges.end(); it++)
				{
					if (it->used_in_loop)
						continue;
					else
					{
						if (it->v1 == node_to_search || it->v2 == node_to_search)
						{
							int edge_start = node_to_search;
							node_to_search = (it->v1 == node_to_search? it->v2 : it->v1);
							prox[i].loop.push_back(edge_start);
							it->used_in_loop = true;
						}
					}	
				}
			}
			// end search a loop for a set of edges
		}
	}
}

void Remesh::orient_loops()
{
	for (int i=0; i<valid_clusters; i++)
	{
		if (prox[i].edges.size() >= 3)
		{			
			vec3 prox_centroid = vec3(0, 0, 0);
			for (vector<int>::iterator it = prox[i].loop.begin(); it != prox[i].loop.end(); it++)
			{
				prox_centroid += nodes[*it];
				prox[i].prox_avg_n += nodes_n[*it];
			}
			prox_centroid /= prox[i].loop.size();
			prox[i].prox_avg_n /= prox[i].loop.size();

			vec3 weighed_n = vec3(0, 0, 0);
			int loop_size = prox[i].loop.size();
			for (int j=0; j<loop_size; j++)
			{
				vec3 v1 = vec3(nodes[prox[i].loop[j]].x, nodes[prox[i].loop[j]].y, nodes[prox[i].loop[j]].z);
				vec3 v2 = vec3(nodes[prox[i].loop[(j+1)%loop_size]].x, nodes[prox[i].loop[(j+1)%loop_size]].y, nodes[prox[i].loop[(j+1)%loop_size]].z);
				weighed_n += (v1-prox_centroid).cross(v2-v1);
			}

			//reverse the loop if needed
			if (weighed_n.dot(prox[i].prox_avg_n) < 0)
				reverse(prox[i].loop.begin(), prox[i].loop.end());
			poly_num++;
		}
	}
}

void Remesh::output_poly_mesh()
{
	string save_file = out_file.substr(0, out_file.find_last_of(".")) + "_simplified_poly.off";
	cout << "output simplified poly mesh to " << save_file.c_str() << endl;
	ofstream of(save_file.c_str());
	of << "OFF" << endl;
	of << nodes.size() << '\t' << poly_num << '\t' << 0 << endl;
	for (int i=0; i< nodes.size(); i++)
	{
		of <<  nodes[i].x << '\t' <<  nodes[i].y << '\t' <<  nodes[i].z << endl;
	}
	for(int i=0; i<valid_clusters; i++)
	{
		if (!prox[i].loop.empty())
		{
			of << prox[i].loop.size();
			for (int j=0; j< prox[i].loop.size(); j++)
			{
				of << '\t' << prox[i].loop[j];
			}
			of << endl;
		}
	}
	of.close();
	cout << "export poly mesh complete!" << endl;
}

void Remesh::construct_CDT_mesh()
{
	dijkstra = new Dijkstra(valid_clusters, init_mesh.mesh_node.size());
	dijkstra->initialize_neighbors();
	for (int i=0; i<valid_clusters; i++)
	{
		vector<pair<int,int>> source_and_scalars;
		for (vector<Anchor_node>::iterator it = prox[i].neighboring_cluster_info.begin(); it != prox[i].neighboring_cluster_info.end(); it++)
		{
			source_and_scalars.push_back(pair<int,int>(it->closest_node, it->node_id));
		}
		dijkstra->initialize_boundary_flooding(i, source_and_scalars);
		dijkstra->flood_cluster_boundary(i);
		dijkstra->flood_cluster_inner(i);
	}
	extract_all_triangles();
	output_CDT_mesh();
}

void Remesh::extract_all_triangles()
{
	set<set<int>> existing_vs;
	Vector3<long int> v;
	for (int i=0; i<init_mesh.mesh_face.size(); i++)
	{
		set<int> s;	
		for (int j = 0; j<3; j++)
		{
			int node_id = init_mesh.mesh_face[i][j];
			s.insert(dijkstra->nearest_scalar[node_id]);
			v[j] = dijkstra->nearest_scalar[node_id];
		}
		if (s.size() == 3 && existing_vs.find(s) == existing_vs.end())
		{
			existing_vs.insert(s);
			vec3 p[3];
			for (int j=0; j<3; j++)
			{
				p[j] = nodes[v[j]];
			}
			vec3 triangle_n = (p[1]-p[0]).cross(p[2]-p[1]);
			int cluster_id = mesh_partition->find_cluster_id(pcluster->cluster_belong[i]);
			if (triangle_n.dot(prox[cluster_id].prox_avg_n) < 0)
			{
				Vector3<long int> vt = Vector3<long int>(v.x, v.z, v.y);
				v = vt;
			}
			triangle_faces.push_back(v);
		}
	}
}

void Remesh::output_CDT_mesh()
{
	string save_file = out_file.substr(0, out_file.find_last_of(".")) + "_simplified_CDT.off";
	cout << "output simplified CDT mesh to " << save_file.c_str() << endl;
	ofstream of(save_file.c_str());
	of << "OFF" << endl;
	of << nodes.size() << '\t' << triangle_faces.size() << '\t' << 0 << endl;
	for (int i=0; i< nodes.size(); i++)
	{
		of <<  nodes[i].x << '\t' <<  nodes[i].y << '\t' <<  nodes[i].z << endl;
	}
	for (vector<Vector3<long int>>::iterator it = triangle_faces.begin(); it != triangle_faces.end(); it++)
	{
		of << 3 << '\t' << it->x << '\t' << it->y << '\t' << it->z << endl;
	}
	of.close();
	cout << "export CDT mesh complete!" << endl;
}
